## This code is written by Davide Albanese, <albanese@fbk.eu>
## (C) 2010 Fondazione Bruno Kessler - Via Santa Croce 77, 38100 Trento, ITALY.

## This program is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.

## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

## You should have received a copy of the GNU General Public License
## along with this program.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ["ElasticNet"]

import elasticnet as en
import numpy as np


class ElasticNet(object):
    """Elastic Net
    
    Modified Landweber iterative procedure implementation.
    """

    def __init__(self, tau, mu, tol=0.01, maxloops=10000):
        """Initialization.
          
        :Parameters:
          tau : float
            L1 parameter
          mu : float
            L2 parameter
          tol : float
            convergence tolerance
          maxloops : integer
            maximum number of iterations          
          """
        
        self.__tau = tau
        self.__mu = mu
        self.__tol = tol
        self.__maxloops = maxloops

        self.__const = 1.1

    def learn(self, x, y, beta_start=None):
        """Compute the regression coefficients.

        :Parameters:
          x : numpy 2d array (n x p)
            matrix of regressors
          y : numpy 1d array (n)
            response
          beta_start : numpy 1d array (p)
            None: zero vector
        """

        if not isinstance(x, np.ndarray):
            raise ValueError("x must be an numpy 2d array")

        if not isinstance(y, np.ndarray):
            raise ValueError("y must be an numpy 1d array")
            
        if x.ndim > 2:
            raise ValueError("x must be an 2d array")
        
        if x.shape[0] != y.shape[0]:
            raise ValueError("x and y are not aligned")

        xc = x - np.mean(x, axis=0)
        yc = y - np.mean(y)

        if beta_start == None:
            beta_start = np.zeros(xc.shape[1], dtype=np.float)
            
        self.__C = self.__const * np.linalg.norm(np.dot(xc, xc.T), ord=2) * 0.5
                
        tau_m = self.__tau * x.shape[0] / self.__C
        mu_m = self.__mu * x.shape[0] / self.__C  # mu_m = self.__mu * tau_m

        self.__omega = np.dot(xc.T, yc)
        
        self.__beta, self.__steps = \
            en.elasticnet(xc, self.__omega, tau_m, mu_m, beta_start,
                          self.__C, self.__tol, self.__maxloops)
    
    def pred(self, x):
        """Compute the predicted response.
        
        :Parameters:
          x : numpy 2d array (nxp)
            matrix of regressors
        
        :Returns:
          yp : 1d ndarray
             predicted response
        """
        
        if not isinstance(x, np.ndarray):
            raise ValueError("x must be an numpy 2d array")

        if x.ndim > 2:
            raise ValueError("x must be an 2d array")

        if x.shape[1] != self.__beta.shape[0]:
            raise ValueError("x and beta are not aligned")
                
        return np.dot(x, self.__beta)
        
    def beta(self):
        """Return b_1, ..., b_p.
        """
        
        return self.__beta
        
    def selected(self):
        """Returns the selected regressors.
        """

        return np.where(self.__beta != 0.0)[0]

    def steps(self):
        """Return the number of steps performed.
        """

        return self.__steps

    
